#include "base.h"
#include "proc.h"
#include "gui.h"
#include "tree.h"
#include "parser.h"

static int spacelist[] = {USER, PID, CPU, MEM, TTY_WIDTH, NICE, STAT, TIME_LENGTH + 2, 10};
static char *labellist[] = {"USER", "PID", "CPU", "MEM", "TTY", "NICE", "STAT", "START TIME", 
		       " COMMAND", NULL};


static char *help_txt[] =
  {
    "[12345678]..sort by corresponding column",
    "0...........tree view",
    "u/d.........move highlight up/down",
    "U/D.........move highligh one page up/down",
    "-/=.........scroll left/right",
    "_/+.........scroll left/right by one page",
    "r...........refresh",
    "k...........kill processes",
    "t...........toggle kill signal",
    "q...........quit",
    "e...........expand/collapse all",
    "ENTER.......expand/collapse",
    "c...........issue command",
    "a...........select all",
    "C...........clear all selections",
    "R...........reverse selections",
    "m...........monitor selected processes",
    "ctrl-g......cancel",
    "f...........find",
    "g...........find again",
    "s...........search and select",
    "\n",    
    NULL
  };


static void murder(int selected, procinfo_t * list, int count, int sig)
{
    int i;
    int group = 0;

    for (i=0; i<count; ++i) {
	if (list[i].is_selected) {
	    if (kill(list[i].proc->tid, sig) == -1) {
	      /* XXX we should do something about... */
	    };
	    group = 1;
	}
    }

    if (!group)
	kill(list[selected].proc->tid, sig);
	
    sleep(1);
}

static void make_item(char *item, procinfo_t procinfo, int highlight)
{

  int i, j, col;
  char *tmp;

  for (i = 0; i < COLS; i++)
    item[i] = ' ';
  
  item[COLS] = '\0';
  
  col = START_COL;

  /****************************************/
  if(field_op_[FUSER]) {
  
  for(i = 0; i < USER; i++) {
    if(procinfo.proc->euser[i] && ((procinfo.proc->euser)[i] != '\0'))
      item[i + col] = (procinfo.proc->euser)[i];
    else
      break;
  }
  col += USER;
  }
  /*---------------------------------------*/

  if(field_op_[FPID]) {
  tmp = (char *) malloc(PID * sizeof(char));
  sprintf(tmp, "%i", procinfo.proc->tid);

  for(i = 0; i < PID; i++) {
    if (tmp[i] && (tmp[i] != '\0'))
      item[i + col] = tmp[i];
    else 
      break;
  }

  free(tmp);
 
  col += PID;
  }
  /*--------------------------------------*/
  
  if(field_op_[FCPU]) {
  tmp = (char *) malloc(CPU * sizeof(char));
  sprintf(tmp, "%2u.%u", procinfo.proc->pcpu / 10, procinfo.proc->pcpu % 10);
  
  for(i = 0; i < CPU; i++) {
    if(tmp[i] && (tmp[i] != '\0'))
      item[i + col] = (char) tmp[i];
    else
      break;
  }

  free(tmp);

  col += CPU;
  }
  /*---------------------------------------*/
  
  if(field_op_[FMEM]) {
  tmp = (char *) malloc(MEM * sizeof(char));
  sprintf(tmp, "%2ld.%ld", procinfo.pmem / 10, procinfo.pmem % 10);
  
  for(i = 0; i < MEM; i++) {
    if(tmp[i]  && (tmp[i] != '\0'))
      item[i + col] = (char) tmp[i];
    else
      break;
  }

  free(tmp);

  col += MEM;
  }
  /******************************************/
  if(field_op_[FTTY]) {
  for(i = 0; i < TTY_WIDTH; i++) {
    if(procinfo.ttyc[i] && (procinfo.ttyc[i] != '\0'))
      item[i + col] = procinfo.ttyc[i];
    else
      break;
  }  
  
  col += TTY_WIDTH;
  }
  /*---------------------------------------*/

  if(field_op_[FNICE]) {
  tmp = (char *) malloc(NICE * sizeof(char));
  sprintf(tmp, "%2ld", procinfo.proc->nice);

  for(i = 0; i < NICE; i++) {
    if(tmp[i]  && (tmp[i] != '\0'))
      item[i + col] = tmp[i];
    else
      break;    
  }

  free(tmp); 
 
  col += NICE;
  }
  /*----------------------------------------*/
  
  if(field_op_[FSTAT]) {
  item[col] = procinfo.proc->state;
  
  col += STAT;
  }
  
  /******************************************/


  if(field_op_[FTIME]) {
    /*for(i = 0; i < 2; i++) {
      item[i + col] = procinfo.etime.hh[i];
    }  
    
    item[col + 2] = ':';
    
    for(j = 0; j < 2; j++) {
      item[3 + j + col] = procinfo.etime.mm[j];
    }
    
    col += TIME;*/
    
    for(i = 0; i < 42 && procinfo.stime[i] != '\0'; i++) {
      item[i + col] = procinfo.stime[i];
    }
    col += TIME_LENGTH;
  }
  /*----------------------------------------*/
  
  
  if(procinfo.is_selected)
    item[col] = '*';

  if(highlight)
    item[col + 1] = '>';

  col += 3;

  /*----------------------------------------*/
  
  for(i = 0; i + col  < COLS; i++) {
    if (procinfo.s_pname[i] && (procinfo.s_pname[i] != '\0'))
      item[i + col] = (procinfo.s_pname)[i];
    else
      break;
  }
  /*-----------------------------------------*/

  item[COLS] = '\0';

}

static void row_clear(int current_row)
{
    int i;
    char *tmp;

    tmp = (char *) malloc((COLS + 1) * sizeof(char));

    attrset(0);

    for (i = 0; i < COLS; i++) {
	tmp[i] = ' ';
    }

    tmp[COLS] = '\0';

    mvprintw(START_ROW + ROW_SPACE + current_row, 0, "%s", tmp);
    free(tmp);
}


static void draw_menu(killer_t *chow, int current_highlight, unsigned int selected, size_t *selected_col, int count, int t_view)
{
    int current_row = 0;
    int i;
    int pos;
    int col = START_COL;
    int row;
    char *str_buf;
    procinfo_t *options;
    int is_selected;
    char *item, *blank;
    char *pname;
    int l;

    if (t_view)
	options = chow->trunk;
    else
	options = chow->pidlist;

    str_buf = (char *) malloc((COLS + 1) * sizeof(char));
    /*  setpwent(); */
    /* printf("in draw menu\n"); */

    blank = (char *) malloc((COLS + 1) * sizeof(char));
    for (i = 0; i < COLS; i++)
	blank[i] = ' ';
    blank[COLS] = '\0';

    for (i = 0; i <= MAX_ROW; i++) {
	row_clear(i);
    }

    if (count) {
	while (current_row < MAX_ROW && selected + current_row - current_highlight < count) {

	    row = START_ROW + ROW_SPACE + current_row;
	    pos = (selected + current_row - current_highlight);
	    row_clear(current_row);

	    if (current_row < count) {
		is_selected = options[pos].is_selected;

		if (is_selected) {
		    attrset(COLOR_PAIR(SELECTED_COLOR) | A_BOLD);
		} else
		    attrset(COLOR_PAIR(NORMAL_COLOR) | A_NORMAL);

		if (current_row == current_highlight) {
		  pname = chow->pidlist[pid_pos(options[pos].proc->tid, chow->pidlist, chow->count)].pname;
		  l = strlen(pname);
		  
		  if(COLS < l) {
		    if(*selected_col > l - COLS + 5)
		      (*selected_col) = l - COLS + 5;
		    strncpy(str_buf, pname + *selected_col, COLS);
		  }
		  else
		    strncpy(str_buf, pname, COLS);		    
		  
		  if (is_selected)
			attrset(COLOR_PAIR(SELECTED_HIGHLIGHT_COLOR) | A_STANDOUT);
		    else
			attrset(COLOR_PAIR(HIGHLIGHT_COLOR) | A_STANDOUT);
		}

		mvprintw(row, 0, "%s", blank);

		item = (char *) malloc((COLS + 1) * sizeof(char));
		
		make_item(item,  options[pos], 
			  (current_row == current_highlight) ? 1 : 0);
		
		mvprintw(row, col, "%s", item);
		
		free(item);
	    }
	    current_row++;
	}
    }	
		       
    if (count) {
	attrset(COLOR_PAIR(BANNER_COLOR) | A_STANDOUT | A_BOLD);
	mvprintw(LINES - 1, 0, "%s", blank);
	mvprintw(LINES - 1, 0, "%s", str_buf);
    }
    refresh();
    free(blank);
}

static void draw_top(int key)
{
  int col = START_COL;
  char *tmp;
  int i, j;
  
  int offset;
  static int tag = 0;
  
  if(key)
    tag = key;
  
  tmp = (char *) malloc(COLS * sizeof(char) + 1);
  for (i = 0; i < COLS; i++) {
    tmp[i] = ' ';
  }
  
  tmp[i] = '\0';

  offset = 0;

  for(i = 1; i < FIELDS; i++) {
    if(field_op_[i]) {
      for(j = 0; j < strlen(labellist[i - 1]); j++) {
	tmp[offset + j] = labellist[i - 1][j];
      }
      offset += (size_t) spacelist[i - 1];
    }
  }

  attrset(COLOR_PAIR(BANNER_COLOR) | A_STANDOUT | A_BOLD);
  mvprintw(START_ROW, START_COL, tmp);

  free(tmp);
  
  if(tag) {

    if(tag < FIELDS)
      j = spacelist[tag - 1];
    else
      j = strlen(labellist[FIELDS - 2]) + 1;
    
    tmp = (char *) malloc(j * sizeof(char));
    
    for (i = 0; i < j; i++) {
      tmp[i] = ' ';
    }
    
    tmp[i] = '\0';
    
    for(i = 0; i < strlen(labellist[tag - 1]); i++) {
      tmp[i] = labellist[tag - 1][i];
    }

    col = START_COL;

    for(i = 1; i < tag; i++) {
      if(field_op_[i]) {
	col += spacelist[i - 1];
      }
    }
    
    attrset(COLOR_PAIR(BANNER_SELECTED_COLOR) | A_BOLD);
    mvprintw(START_ROW, col, tmp);
  
    free(tmp);
  }

}


static int help_screen(void)
{
  int i;
  int row = 2;
  int line = 0;
  int punch;
  int check = 1;
  char*blank;

    blank = (char *) malloc((COLS + 1) * sizeof(char));

    for(i = 0; i < COLS; i++) {
      blank[i] = ' ';
    }
    
    blank[COLS] = '\0';

    clear_all_screen();
    
    attrset(COLOR_PAIR(BANNER_COLOR) | A_STANDOUT | A_BOLD);

    //attrset(COLOR_PAIR(NORMAL_COLOR));
    mvprintw(START_ROW, 0, "%s", blank);
    mvprintw(START_ROW, 4, "press 'q' to exit this screen");

    while(help_txt[line]) {
      if(row < LINES - 2) {
	attrset(COLOR_PAIR(NORMAL_COLOR));
	mvprintw(row, 4, "%s", help_txt[line]);
	line++;
	row++;
	} else {
	  attrset(COLOR_PAIR(BANNER_COLOR) | A_STANDOUT | A_BOLD);
	  mvprintw(LINES - 1, 0, "%s", blank);
	  mvprintw(LINES - 1, 4, "press space to go to next page");
	  while(check) {
	    fflush(stdin);
	    punch = getch();
	    if(punch == ' ') {
	      clear_all_screen();
	      attrset(COLOR_PAIR(BANNER_COLOR) | A_STANDOUT | A_BOLD);
	      mvprintw(0, 0, "%s", blank);
	      mvprintw(0, 4, "press 'q' to exit this screen");
	      row = 2;
	      check = 0;
	    } else if (punch == 'q')
	      return 1;
	  }
	}
    }

    attrset(COLOR_PAIR(BANNER_COLOR) | A_STANDOUT | A_BOLD);
    mvprintw(LINES - 1, 0, "%s", blank);

    check = 1;

    while(check) {
      fflush(stdin);
      punch = getch();
      if(punch == 'q') {
	free(blank);
	return 1;
      }
    }

    return 1;

}


static void input_dlg(char *input, char *greet)
{
    WINDOW *dialogue;
    int i, check, pos, length, j;
    /* char blank[COLS]; */ /* gcc doent like */

    char *blank;
    
    int tmp;
    int field = 0;
    int init_pos = strlen(greet) + 1;

    blank = (char * ) malloc(COLS * sizeof(char));
    
    for (i = 0; i < COLS - 1; i++)
	blank[i] = ' ';

    blank[i] = '\0';

    dialogue = newwin(1, COLS, LINES - 1, 0);

    wbkgd(dialogue, COLOR_PAIR(0));

   
    wattrset(dialogue, COLOR_PAIR(0));

    mvwprintw(dialogue, 0, 0, "%s", greet);
    
    noecho();
    curs_set(1);
    refresh();

    wrefresh(dialogue);
    /*  mvwgetstr(dialogue, 1, 0, input); */
    /*  wscanw(dialogue, "which da fuck one? %s", &input); */

    length = strlen(input);;
    i = 0;
    check = 1;
    pos = init_pos;

    while (check) {
	mvwprintw(dialogue, 0, 0, "%s", greet);
	if (field == 0) {
	    input[length] = '\0';
	    wattrset(dialogue, COLOR_PAIR(0));
	    mvwprintw(dialogue, 0, init_pos, "%s", blank);
	    mvwprintw(dialogue, 0, init_pos, "%s", input);
	   
	    wmove(dialogue, 0, pos);
	    wrefresh(dialogue);
	    /*  mvwprintw(dialogue, 3, pos, "%s", "    "); */
	    /*  tmp = mvwgetch(dialogue, 2, pos); */
	    tmp = getch();
	    if (tmp != ctrl ('G')) {
		if (tmp != '\n') {
		    if (tmp == KEY_BACKSPACE) {
			if (pos > init_pos) {
			    for (j = i; j < length; j++) {
				input[j - 1] = input[j];
			    }
			    pos--;
			    i--;
			    length--;
			}
		    } else if (tmp == KEY_LEFT) {
			if (i > 0) {
			    pos--;
			    i--;
			}
		    } else if (tmp == KEY_RIGHT) {
			if (i < length) {
			    pos++;
			    i++;
			}
		    } else if (tmp == ' ') {
			if (length < BOX_WIDTH - 5) {
			    for (j = length; j > i; j--) {
				input[j] = input[j - 1];
			    }
			    input[i] = ' ';
			    pos++;
			    length++;
			    i++;
			}
		    } else if (tmp == KEY_DC) {
			if (i < length) {
			    for (j = i; j < length; j++) {
				input[j] = input[j + 1];
			    }
			    length--;
			}
		    } else {
			if (length < BOX_WIDTH - 5) {
			    for (j = length; j > i; j--) {
				input[j] = input[j - 1];
			    }
			    input[i] = tmp;
			    pos++;
			    length++;
			    i++;
			}
		    }
		} else
		    check = 0;
	    } else {
		field = (field + 1) % 2;
	    }
	} else {
	    /*  mvwprintw(dialogue, 3, pos, "%s", "    "); */
	  input[0] = '\0';
	  check = 0;
	}
    }

    input[length] = '\0';
   
    wattrset(dialogue, COLOR_PAIR(0));
    
    wrefresh(dialogue);
    noecho();
    curs_set(0);
    /*  tmp = mvwgetch(dialogue, 5, 2); */
    delwin(dialogue);
    refresh();
    free(blank);

}

static void move_up(int times, int *selected_row, unsigned int *selected)
{
    if(*selected_row >= times)
	*selected_row -= times;
    else
	*selected_row = 0;

    if(*selected >= times)
	*selected -= times;
    else
	*selected = 0;
    
}


static void move_down(int times, int *selected_row, unsigned int *selected, int count)
{
    if(times <= 0)
      return;

    if (count) {
	if (*selected + times <= count - 1) 
	    (*selected) +=  times;
	else
	    *selected = count - 1;
	    
	if (*selected_row + times <= (MAX_ROW - 1) && *selected_row + times <= count - 1) {
	    (*selected_row) += times;
	}
	else {
	    if (MAX_ROW - 1 < count - 1)
		*selected_row = (MAX_ROW - 1);
	    else
		*selected_row = count - 1;
	}
	
    }
}

int getchoice(killer_t * chow)
{
    static int selected_row = 0;
    static unsigned int selected = 0;
    static size_t selected_col = 0;
    
    static int buffer = 0;
    static int key = 0;
    static int tag = 0;
    static int parity = 1;
    
    static int sig = 0;
    #define NRSIGS 6
    int siglist[NRSIGS] = {1, 9, 6, 3, 2, 15};
    char *signame[NRSIGS] = { "SIGHUP ", "SIGKILL", "SIGABRT", 
			 "SIGQUIT", "SIGINT ", "SIGTERM" };
    int tmp = NORMAL;
    char sigstr[8];
    procinfo_t *list;
    static int e_view = 1;

    int i;

    int count;
    static char *finput;
    char *input;
     
    int pos;

    killer_t *sub;    
     
    input = (char *) malloc((BOX_WIDTH - 5) * sizeof(char));

    strcpy(sigstr, signame[sig]);  

    if (key == 'k' || key == 'r' || key == '\n' || key == 'e' || key == ' '
	|| key == 'c' || (key >= '0' && key <= '0' + fields_)) {
	
      if (key >= '1' && key <= '0' + fields_ ) {
	  if(key == tag)
		parity *= -1;
	  tag =  key;
      }
      shell_sort(chow->pidlist, parity * avfields_[tag - '0' - 1], chow->count);
      
      if(key ==  '0') {
	for (i = 0; i <= MAX_ROW; i++) {
	  row_clear(i);
	}
	if (t_view == 0) {
	  /* 	 selected_row = selected = 0; */
	  t_view = 1;
	} else
	  t_view = 0;
      }
	  
      if(t_view)
	make_tree(chow);

      if (e_view) {
	for (i = 0; i < chow->count; i++) {
	  chow->pidlist[i].expand = 1;
	}
      }	
      
      chow->t_count = 0;
      
      graph_branch(chow->ptree, chow, 0);
      
      draw_strings(chow->trunk, chow->t_count);
    
    }
    
    if (key != KEY_LEFT && key != '-' && key != KEY_RIGHT && key != '='
	&& key != KEY_HOME && key != '_' && key != KEY_END && key != '+')
      selected_col = 0;

    if (t_view) {
      list = chow->trunk;
      count = chow->t_count;
    } else {
      list = chow->pidlist;
      count = chow->count;
    }

    if(selected > count - 1) {
      if (count - 1 < MAX_ROW) {
	selected = selected_row = count - 1;
      } else {
	selected = selected_row = 0;
	move_down(buffer, &selected_row, &selected, count);
      }
    }

#if !DEBUG
    if(tag >= '1' && tag <= '0' + fields_)
	draw_top(avfields_[tag - '0' - 1]);
    else
	draw_top(0);
    attrset(COLOR_PAIR(NORMAL_COLOR));
    mvprintw(0, COLS - strlen(sigstr), "%s", sigstr);
    //refresh();
#endif

#if !DEBUG
    draw_menu(chow, selected_row, selected, &selected_col, count, t_view);
#endif	
          
    fflush(stdin);
    key = getch();

    switch(key) {

    case KEY_LEFT:
      if(selected_col > 0)
	 selected_col--;
       break;
      
    case '-':
      if(selected_col > 0)
	selected_col--;
      break;
      
    case KEY_HOME:
      if(selected_col > COLS)
	selected_col -= COLS;
      else
	selected_col = 0;
      break;

    case '_':
      if(selected_col > COLS)
	selected_col -= COLS;
      else
	selected_col = 0;
      break;
      
    case KEY_RIGHT:
      selected_col++;
      break;

    case '=':
      selected_col++;
      break;

    case KEY_END:
      selected_col += COLS;
      break;
      
    case '+':
      selected_col += COLS;
      break;
      
    case KEY_UP:
      move_up(1, &selected_row, &selected);
      break;
      
    case 'u':
      move_up(1, &selected_row, &selected);
      break;

    case  KEY_DOWN:
      move_down(1, &selected_row, &selected, count);
      break;
      
    case  'd':
      move_down(1, &selected_row, &selected, count);
      break;

    case KEY_NPAGE:
      move_down(MAX_ROW, &selected_row, &selected, count);
      break;
      
    case 'D':
      move_down(MAX_ROW, &selected_row, &selected, count);
      break;

    case KEY_PPAGE:
      move_up(MAX_ROW, &selected_row, &selected);
      break;

    case 'U':
      move_up(MAX_ROW, &selected_row, &selected);
      break;

    case 'k':
      if(count)
	murder(selected, list, count, siglist[sig]);
      tmp = REINIT;
      break;
     
    case '\n':
      if (t_view) {
	pos = pid_pos(list[selected].proc->tid, chow->pidlist, chow->count);
	chow->pidlist[pos].expand = (chow->pidlist[pos].expand + 1) % 2;
	e_view = 0;
	clear_all_screen();
      }
	tmp = NORMAL;	    
	break;

    case 'e':
      for (i = 0; i <= MAX_ROW; i++) {
	row_clear(i);
      }
      if (t_view) {
	if (!e_view) {
	  for (i = 0; i < chow->count; i++) {
	    chow->pidlist[i].expand = 1;
	  }
	  e_view = 1;
	} else
	  for (i = 0; i < chow->count; i++) {
	    chow->pidlist[i].expand = 0;
	    e_view = 0;
	  }
	clear_all_screen();
      }
      tmp = NORMAL;
      break;

    case 's':
      if (chow->count) {
	input[0] = '\0';
	input_dlg(input, "search and select:");
	if (input[0] != '\0') {
	  search(chow, input, t_view);
	}
      }
      tmp = NORMAL;
      break;	
    
    case 'f':
      if (chow->count){	      
	if(finput)
	  free(finput);
	finput = (char * ) malloc((BOX_WIDTH - 5) * sizeof(char));
	finput[0] = '\0';
	input_dlg(finput, "find:");
	if (finput[0] != '\0') {
	  move_down(find(chow, selected, finput, t_view) -  selected, &selected_row, &selected, chow->count);
	}
      }
      tmp = NORMAL;
      break;

    case  'g':
      if (chow->count && finput) {
	if (finput[0] != '\0') {

	  move_down(find(chow, selected, finput, t_view) -  selected, &selected_row, &selected, chow->count);
	}
      }
      tmp = NORMAL;
      break;

    case 'c':
      if (chow->count) {
	sprintf(input, "%i", list[selected].proc->tid);
	input_dlg(input, "command:");
	if (input[0] != '\0') {
	  def_prog_mode();
	  system("clear");
	  system(input);
	  system("echo '\npress any key to continue.'");
	  getchar();
	  reset_prog_mode();
	  tmp = REINIT;
	} else
	    tmp = NORMAL;
      } else
	tmp = NORMAL;
      break;
      
    case ' ':
      if (count) {
	sub = chow;
	pos = pid_pos(list[selected].proc->tid, sub->pidlist, sub->count);
	sub->pidlist[pos].is_selected = (sub->pidlist[pos].is_selected + 1) % 2;
      }
      move_down(1, &selected_row, &selected, count);
      tmp = NORMAL;
      break;
      
    case 'a':
      for (i = 0; i < count; i++) {
	sub = chow;
	pos = pid_pos(list[i].proc->tid, sub->pidlist, sub->count);
	sub->pidlist[pos].is_selected = 1;
      }
      tmp = NORMAL;
      break;
      
    case 'C':
      for (i = 0; i < chow->count; i++) {
	chow->pidlist[i].is_selected = 0;
      }
      tmp = NORMAL;
      break;
      
    case 'R':
      for (i = 0; i < count; i++) {
	int pmt;
	sub = chow;
	  pos = pid_pos(list[i].proc->tid, sub->pidlist, sub->count);
	  pmt = sub->pidlist[pos].is_selected;
	  sub->pidlist[pos].is_selected = !pmt;
      }
	tmp = NORMAL;
	break;
	
    case 'q':
      tmp = QUIT;
      break;
      
    case 'r': 
      tmp = REINIT;
      break;
    
    case 'p':
      tmp = LEFT;
      break;
      
    case 't':
      sig = (sig + 1) % NRSIGS;
      break;
    
    case 'h':
      help_screen();
      tmp = NORMAL;
      break;
      
    default:
      tmp = NORMAL;
 
    }
    
    free(input);

    buffer = selected;

    return tmp;

}


void clear_all_screen()
{
    clear();
    refresh();
}

void check_scr_size()
{
  int i, width;
  
  width = 0;
  
  for(i = 1; i < FIELDS; i++){
    if (field_op_[i])
      width += spacelist[i - 1];
  }
  
    if(COLS < width) {
	clear_all_screen();
	endwin();

	printf("error: terminal size too small\n");
	exit(0);
    }
}
