#include "base.h"
#include "parser.h"

static char *color_template[] = 
  {
    "normal_color",
    "selected_color",
    "banner_color",
    "highlight_color",
    "selected_highlight_color",
    "banner_selected_color",
    NULL
  };

static int read_line(FILE *fp, char *s) 
{

  char *ch;  
  
  if(fgets(s, STRING_LENGTH, fp) == NULL)
    return 0;

  if(s[0] == '#')
    return 2;
  
  if ((ch = strchr(s, '\n')) != NULL)
    {
      return 1;
    } 
  else 
    {
      int c;
      c = getc (fp); /* This is kind of a hack. We want to know if the
                        char at the current point in the input stream is EOF.
                        feof() will only tell us if we've already hit EOF, not
                        if the next character is EOF. So, we need to read in
                        the next character and manually check if it is EOF. courtesy of mutt*/
      if (c == EOF)
      {
        return 1;
      }
      
    }

  return 0;
  
}
 
static void parse_line_color(color_config_t *color_config, char *s)
{

  int i = 1;
  size_t offset = 0;
  size_t j = 0;
  int check = 1;
  char ch;
  int space;
  
  space = 0;
  while(s[space] == ' ' || s[space] == '\t')
    space++;

  offset += (size_t) space;

  while(check && i < COLOR_OPTIONS) {
	if(strncmp(color_template[i - 1], s + offset, strlen(color_template[i - 1])) == 0) {
	  check = 0;
	}
    i++;
  }

  offset +=  strlen(color_template[i - 2]) - 1;

  if(!check) {
    for(j = 0; (int) j + (int) offset < strlen(s); j++) {
      ch = (s + offset + j)[0];
      if(ch == '\n' || ch == EOF)
	break;
      if (ch >= '0' && ch < '8') {
	color_config[i - 1].background = ch % '0';
	break;
      }
    }

    offset = offset + j + 1;
    
    for(j = 0; (int) j + (int) offset < strlen(s); j++) {
      ch = (s + offset + j)[0];
      if(ch == '\n' || ch == EOF)
	break;      
      if (ch >= '0' && ch < '8') {
	color_config[i - 1].foreground = ch % '0';
	break;
      }
    }
  }
}
    
int *parse_usernames(char *users, int *uid)
{

  char name_buf[17];
  int i, l;
  int offset = 0;
  int increment = 5;
  int vacancy = 5;
  int listsize = 1;
  struct passwd *p;

  l = strlen(users);

  uid = (int *) malloc(vacancy * sizeof(int));

  while(offset < l) {
    i = 0;
    while((users[i + offset] != ',') && (i + offset < l)) {
      if (i > 17) {
	fprintf(stderr, "ncps: username too long.");
	exit(1);
      }
      else {
	name_buf[i] = users[i + offset];
	i++;
      }
    }

    name_buf[i] = '\0';

    offset = i + offset + 1;

    p = getpwnam(name_buf);
    if (p == NULL) {
      fprintf(stderr, "ncps: No such user: %s\n", name_buf);
      exit(1);
    } else {
      if (listsize > vacancy - 1) {
	vacancy += increment;
	uid = (int *) realloc(uid, vacancy * sizeof(int));
      }
      uid[listsize] = (int) p->pw_uid;
      listsize++;
    }
  }
   
  uid[0] = listsize - 1;

  return uid;

}


int parse_rc_color(const char *path, color_config_t *color_config)
{
  
  FILE *f;
  char *linebuf; 
  int i;

  color_config[0].background = COLOR_BLACK;
  color_config[0].foreground = COLOR_WHITE;
  
  color_config[NORMAL_COLOR].background = COLOR_BLACK;
  color_config[NORMAL_COLOR].foreground = COLOR_WHITE;

  color_config[SELECTED_COLOR].background = COLOR_RED;
  color_config[SELECTED_COLOR].foreground = COLOR_GREEN;

  color_config[BANNER_COLOR].background = COLOR_GREEN;
  color_config[BANNER_COLOR].foreground = COLOR_BLUE;

  color_config[HIGHLIGHT_COLOR].background = COLOR_BLACK;
  color_config[HIGHLIGHT_COLOR].foreground = COLOR_CYAN;

  color_config[SELECTED_HIGHLIGHT_COLOR].background = COLOR_BLACK;
  color_config[SELECTED_HIGHLIGHT_COLOR].foreground = COLOR_RED;

  color_config[BANNER_SELECTED_COLOR].background = COLOR_BLACK;
  color_config[BANNER_SELECTED_COLOR].foreground = COLOR_WHITE;

  f = fopen(path, "r");

  if(f == NULL) {
    return 1;
  }

  linebuf = (char *) malloc(STRING_LENGTH * sizeof(char));
  
  while((i = read_line(f, linebuf))) {
    if(i == 1)
      parse_line_color(color_config, linebuf);
  }

  free(linebuf);
  
  fclose(f);
  
  return 0;
    
}



