#include <stdio.h>
#include <stdlib.h>
#include <regex.h>
#include <time.h>

#include "base.h"
#include "escape.h"
#include "proc.h"

static unsigned long page_shift;
static unsigned long seconds_since_1970;
static unsigned long seconds_since_boot;

static void pr_etime(etime_t *etime, unsigned long start_time)  /* blatantly ripped from ps source code */
{

  unsigned t, dd, hh, mm, ss;
  
  t = (
       ((unsigned long)uptime(0,0))
       - ((unsigned long) start_time)
       / Hertz
       );
  
  ss = t%60;
  t /= 60;
  mm = t%60;
  t /= 60;
  hh = t%24;
  t /= 24;
  dd = t;

  snprintf(etime->dd, 239, "%u-", dd);
  snprintf(etime->hh, 3, "%02u", hh);
  snprintf(etime->mm, 3, "%02u", mm);
  snprintf(etime->ss, 3, "%02u", ss);
}


static void pr_stime(char *s, unsigned long start_time){ /* shamelessly ripped from ps source code */ 
  struct tm *proc_time;
  struct tm *our_time;
  time_t t;
  char *fmt;
  int tm_year;
  int tm_yday;

  our_time = localtime(&seconds_since_1970);   /* not reentrant */
  tm_year = our_time->tm_year;
  tm_yday = our_time->tm_yday;
  t = (time_t)(
            ((unsigned long) (seconds_since_1970 - seconds_since_boot))
          + (start_time)
          / Hertz
  );
  proc_time = localtime(&t); /* not reentrant, this corrupts our_time */
  fmt = "   %H:%M:%S";                                   /* 03:02 23:59 */
  if(tm_yday != proc_time->tm_yday) fmt = "%b%d %H:%M";  /* Jun06 Aug27 */
  if(tm_year != proc_time->tm_year) fmt = "%Y";    /* 1991 2001 */
  strftime(s, TIME_LENGTH, fmt, proc_time);
  
}


void fill_pcpu(proc_t * buf)
{
  unsigned long total_time;
  unsigned long pcpu = 0;
  unsigned long seconds;
 
    if (buf == NULL)
	return;			/* Just in case */

    total_time = buf->utime + buf->stime;

    seconds = uptime(0, 0)  -  ((unsigned long)buf->start_time) / Hertz;

    if (seconds)
	pcpu = (total_time * 1000ULL / Hertz) / seconds;
    buf->pcpu = (pcpu > 999) ? 999 : pcpu;
}

static procinfo_t *simple_spew(killer_t *chow, uid_t *uidlist)
{
    proc_t buf;
    PROCTAB *ptp;
    //int check = 1;
    int i;
    int proc_table_size = PT_SIZE;
    procinfo_t *tmp;

    procinfo_t *tmptmp;

    proc_table_size = PT_SIZE;
    chow->pidlist = malloc(proc_table_size * sizeof(procinfo_t));

    tmp = chow->pidlist;

    if (tmp == NULL) {
	/* XXX we should realy do something about this... */
	exit(-1);
    }
    //ptp = openproc(PROC_FILLCOM | PROC_FILLENV | PROC_FILLUSR | PROC_FILLBUG);
    if (uidlist)
      {
	ptp = openproc (PROC_FILLMEM | PROC_FILLCOM | PROC_FILLENV |  PROC_FILLSTAT | PROC_FILLSTATUS | PROC_FILLUSR | PROC_UID,
		  uidlist + 1, uidlist[0]);
      }
    else
      ptp = openproc (PROC_FILLBUG);
    

    i = 0;
    memset(&buf, '#', sizeof(proc_t));
    while (readproc(ptp, &buf)) {
      //if (buf.suid == uid || uid == -1 {
	    if (i > proc_table_size - 1) {	/* Pajtek */
		proc_table_size += PT_SIZE;
		tmptmp = realloc(tmp, proc_table_size * sizeof(procinfo_t));
		if (tmptmp == NULL) {
		    /* XXX some problems reallocating... */
		    exit(-1);
		}
		tmp = tmptmp;
	    }
	    tmp[i].proc = (proc_t *) malloc(sizeof(proc_t));
	    *(tmp[i].proc) = buf;	/* Pajtek */
	    i++;
	    memset(&buf, '#', sizeof(proc_t));
	
    }

    chow->count = i;

    after_spew(tmp, chow->count);

    closeproc(ptp);
    
#if DEBUG
    for (i = 0; i < chow->count; i++) {
	printf("%i\n", tmp[i].proc->tid);
    }
#endif

#if DEBUG
    printf("loop over\n");
#endif

    if (chow->count)
      return tmp;
    else
      return NULL;
}

void after_spew(procinfo_t *tmp, int count)
{

  int i;

  for(i = 0; i < count; i++) {

#if DEBUG
    printf("as pid: %i\n", tmp[i].proc->tid);
#endif   

    tmp[0].proc->euser[15] = '\0';

#if DEBUG
    printf("euesr: %s\n", tmp[i].proc->euser);
#endif

    tmp[i].pmem = tmp[i].proc->rss * 1000 / PAGEDIV;
    dev_to_tty(tmp[i].ttyc, TTYC, tmp[i].proc->tty,
	       tmp[i].proc->tid, ABBREV_DEV);
    
    /* in tree view, PID 1 is automatically expanded */
    tmp[i].expand = (tmp[i].proc->tid == 1) ? 1 : 0;
    
    tmp[i].is_selected = check_selection(tmp[i].proc->tid);
    
    fill_pcpu(tmp[i].proc);
    get_lc(&(tmp[i]));
    
    tmp[i].s_pname = strdup(tmp[i].pname);
    
    pr_etime(&(tmp[i].etime), tmp[i].proc->start_time);
    pr_stime(tmp[i].stime, tmp[i].proc->start_time);

  }

}

void initplist(killer_t *chow)
{

  uid_t *uidlist;

  seconds_since_1970 = time(NULL);
  seconds_since_boot = uptime(0,0);

#if DEBUG
    int i;

    printf("initplist\n");
#endif

    uidlist = chow->uidlist;

    meminfo();
    switch (getpagesize()) {
    case 65536:
	page_shift = 16;
	break;
    case 32768:
	page_shift = 15;
	break;
    case 16384:
	page_shift = 14;
	break;
    case 8192:
	page_shift = 13;
	break;
    default:
	fprintf(stderr, "Unknown page size! (assume 4096)\n");
    case 4096:
	page_shift = 12;
	break;
    case 2048:
	page_shift = 11;
	break;
    case 1024:
	page_shift = 10;
	break;
    }

    chow->pidlist = simple_spew(chow, uidlist);

#if DEBUG
    printf("spew complete\n");
#endif

    chow->trunk = (procinfo_t *) calloc(chow->count, sizeof(procinfo_t));

#if DEBUG
    printf("chow->count %i\n", chow->count);
    //for (i = 0; i < chow->count; i++) {
//	printf("%s\n", chow->pidlist[i].pname);
  //  }
#endif
}

int find(killer_t *chow, int selected, char *input, int t_view) {
    
    regex_t preg;
    int tmp = selected + 1;
    int check = 1;
    int count;

    procinfo_t *list;

    if(t_view) {
	list = chow->trunk;
	count = chow->t_count;
    }
    else {
      list = chow->pidlist;
      count = chow->count;
    }

    if (regcomp(&preg, input, REG_ICASE | REG_NOSUB | REG_EXTENDED) != 0)
	return 0;

    while(tmp < count && check) {
	if (regexec(&preg, list[tmp].pname, 0, NULL, 0) != 0)
	    tmp++;
	else 
	    check = 0;
    }
    
    if (!check)
      return tmp;
    else
      return 0;
}

void search(killer_t *mother, char *input, int t_view)
{

  regex_t preg; 
    int i;
    int count;
    int pos;

    procinfo_t *list;
    
    if(t_view) {
      list = mother->trunk;
      count = mother->t_count;
    }
    else {
      list = mother->pidlist;
      count = mother->count;
    }
    
    if (regcomp(&preg, input, REG_ICASE | REG_NOSUB | REG_EXTENDED) != 0)
	return;

    for (i = 0; i < count; i++) {
	if (regexec(&preg, list[i].pname, 0, NULL, 0) == 0) {
	  pos = pid_pos(list[i].proc->tid, mother->pidlist, mother->count);
	  mother->pidlist[pos].is_selected = 1;
	}
    }

}

void get_lc (procinfo_t * input)
{
  char tag[OUTBUF_SIZE], buf[64];
  const char **lc = (const char **) input->proc->cmdline;

  if (lc && *lc)
    {
      escape_strlist (tag, lc, OUTBUF_SIZE);
    }
  else
    {
      /* cmd between '[' and ']' */
      escape_str (buf, input->proc->cmd, 64);
      snprintf (tag, strlen (buf) + 3, "%s", buf);
      tag[strlen (buf) + 3] = '\0';
    }

  input->pname = strdup (tag);
#if DEBUG
  printf ("get_lc: %s\n", input->pname);
#endif
}

void monitor(killer_t * mother)
{
    int i;

    mother->son = (killer_t *) malloc(sizeof(killer_t));

    mother->son->pidlist = (procinfo_t *) calloc(mother->count, sizeof(procinfo_t));
    mother->son->trunk = (procinfo_t *) calloc(mother->count, sizeof(procinfo_t));

    mother->son->monitoring = 1;
    mother->son->mother = mother;

    mother->son->count = 0;
    mother->son->t_count = 0;

    for (i = 0; i < mother->count; i++) {
	if (mother->pidlist[i].is_selected) {
	    dup_procinfo_t(&(mother->son->pidlist[mother->son->count]), &(mother->pidlist[i]));
	    mother->son->pidlist[mother->son->count].s_pname = strdup(mother->son->pidlist[mother->son->count].pname);
	    mother->son->count++;
	}
    }
}

int check_selection(int pid)
{

  int i = 0;
  
  while(selection_template[i] != -1) {

    if (pid == selection_template[i])
      return 1;
    else
      i++;
  }

  return 0;

}

void refresh_stemplate(killer_t * mother)
{      
    int i;
    int j = 0;		
 
    template_size = PT_SIZE;
    selection_template = (int * ) malloc(template_size * sizeof(int));
    
      for (i = 0; i < mother->count; i++) {
	if (mother->pidlist[i].is_selected == 1) {

	    if (j > template_size - 2) {  /* Pajtek */
		template_size += PT_SIZE;
		selection_template = realloc(selection_template, template_size * sizeof(int));
	    }
	    selection_template[j] = mother->pidlist[i].proc->tid;
	    j++;
	}
      }
      selection_template[j] = -1;
}

void shell_sort(procinfo_t *pidlist, int crit, int count) {

    int gap;
    int i;
    int swap;
    long diff = 0;
    procinfo_t temp;
    int parity = 1;
    int tag;
    
    if(crit < 0) {
	parity = -1;
	tag = -1 * crit;
    }
    else {
	tag = crit;
	parity = 1;
    }
    
    if(count % 2)
	gap = (count + 1) / 2;
    else
	gap = count / 2;


    while (gap > 0 && tag) {
	swap = 1;
	while(swap == 1) {
	    swap = 0;
	    for(i = 0; i < count - gap; i++) {
		if(tag == FUSER)
		    diff = alpha(pidlist[i].proc->euser, pidlist[i + gap].proc->euser);
		//diff = pidlist[i].proc->euid - pidlist[i + gap].proc->euid;
		else if (tag == FTTY)
		    diff = alpha(pidlist[i].ttyc, pidlist[i + gap].ttyc);
		else if (tag == FPID)
		    diff = pidlist[i].proc->tid - pidlist[i + gap].proc->tid;
		else if (tag == FCPU)
		    diff = pidlist[i].proc->pcpu - pidlist[i + gap].proc->pcpu;
		else if (tag == FMEM)
		    diff = pidlist[i].pmem - pidlist[i + gap].pmem;
		else if (tag == FNICE)
		  diff = pidlist[i].proc->nice - pidlist[i + gap].proc->nice;
		else if (tag == FSTAT)
		  diff =  (pidlist[i].proc->state - pidlist[i + gap].proc->state);
		else if (tag == FTIME)
		  diff = (pidlist[i].proc->start_time - pidlist[i + gap].proc->start_time);
		else if (tag == FNAME)
		  diff = alpha(pidlist[i].pname, pidlist[i + gap].pname);
		else
		  diff = 0;

		if(diff)
		    diff *= parity;
		
		if (diff < 0) {
		    temp = pidlist[i];
		    pidlist[i] = pidlist[i + gap];
		    pidlist[i + gap] = temp;
		    swap = 1;
		} else
		  swap = 0;

	    }
	}
	if(gap % 2 && gap != 1)
	    gap = (gap + 1) / 2;
	else if (gap == 1)
	  gap = 0;
	else
	  gap = gap / 2;
    }

}

int pid_pos(pid_t pid, procinfo_t *pidlist, int count) {

    int i;
    int check;

    check = 1;
    i = 0;

    while(check && i < count) {	
	if (pidlist[i].proc->tid == pid)
	    check = 0;
	else
	    i++;
    }

    if(i < count)
	return i;
    else
	return -1;

}

long alpha(char *string1, char *string2) {

    int check = 1;
    long diff = 0;
    char *s1, *s2;

    s1 = string1;
    s2 = string2;
    
    while(string1 && string2 && string1[0] != '\0' && string2[0] != '\0' && check) {

	if(string1[0] != string2[0]) {
	    check = 0;
	    diff = string2[0] - string1[0];
	} else {
	    string1++;
		string2++;
	}

    }

    string1 = s1;
    string2 = s2;
    
    return diff;

}

void dup_procinfo_t(procinfo_t *target, procinfo_t *orig) {
    
    memcpy(target, orig, sizeof(procinfo_t));
    target->pname = NULL;
    target->s_pname = NULL;
        
    target->proc = (proc_t *) malloc(sizeof(proc_t));
    memset(target->proc, '#', sizeof(proc_t));
    memcpy(target->proc, orig->proc, sizeof(proc_t));
    target->pname = strdup(orig->pname);
    /*  target->s_pname = strdup(orig->s_pname); */
    strcpy(target->proc->suser, orig->proc->suser);
    
    strcpy(target->proc->cmd, orig->proc->cmd);
    strcpy(target->ttyc, orig->ttyc);
    target->proc->environ = NULL;
    target->proc->cmdline = NULL;

}

void free_list(procinfo_t *pidlist, int count) {

    int i;
    
    for(i = 0; i < count; i++) {
	freeproc(pidlist[i].proc);
	if(pidlist[i].pname)
	    free(pidlist[i].pname);
	  	
	if(pidlist[i].s_pname)
	  free(pidlist[i].s_pname);
    }
    
    free(pidlist);
}

